# FlashRNN - RNNs going Brr on GPUs
## Simple use

```python
import torch
from flashrnn import flashrnn

dtype = torch.bfloat16
device = 'cuda'
B, T, H, D = 1, 1024, 1, 1024

Wx = torch.randn([B, T, 4, H, D], device=device, dtype=dtype)  # 4 gates for LSTM
R = torch.randn([4, H, D, D], device=device, dtype=dtype)  # 4 gates for LSTM
b = torch.randn([4, H, D], device=device, dtype=dtype)

# call the alternating kernel
flashrnn(Wx, R, b, backend="cuda", function="lstm")

# call the fused kernel
flashrnn(Wx, R, b, backend="cuda_fused", function="lstm")
```

## Implement your own RNN

1. Add pointwise functions in `flashrnn/vanilla/MYRNN.py`, `flashrnn/alternating/MYRNN_pointwise.cu`, `flashrnn/fused/MYRNN_fused_pointwise.cuh`. 
2. Add pointwise python function to python in `flashrnn/vanilla/__init__.py`.
3. Add MYRNN to rnn_function_registry in `flashrnn/flashrnn.py`
4. Enjoy the speed.

Note that currently the fused kernel only supports gate numbers that are powers of two (that divide 16).